/* GNet - Networking library
 * Copyright (C) 2000  David Helder
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include "gnet-private.h"
#include "url.h"


GURL* 
gnet_url_new (const gchar* url)
{
  GURL* gurl = NULL;
  gchar* local_url;
  const gchar* p;
  const gchar* temp;
  const gchar* inner_temp;
  
  g_return_val_if_fail (url, NULL);

  gurl = g_new0 (GURL, 1);

  /* Strip away whitspaces */
  local_url = g_strdup (url);

  p = g_strstrip (local_url);
  temp = p;
  
  /* Scheme part */
  
  while (*p != 0 && *p != ':' && *p != '/')
    ++p;

  if (*p == ':' && p[1] != 0)
    {
      gurl->protocol = g_strndup (local_url, p - local_url); 
    }
  else
    p = temp;

  /* Authority part */
  
  if ((gurl->protocol != NULL && p[1] == '/' && p[2] == '/') || /* Is net path */
      (gurl->protocol != NULL && p[1] != '/')) /* For mailto: and news: */
    {
      if (p[1] == '/')
        p += 3; /* Skip the "://" part */
      else
        ++p; /* Skip the : */
      
      /* Look for user and password */

      temp = p;

      while (*p != 0 && *p != '@' && *p != '/' )
        ++p;

      if (*p == '@') /* Has user and possibly password */
        {
          inner_temp = p;
          p = temp;
          
          while (*p != 0 && *p != ':' && *p != '@')
            ++p;

          if (*p == ':') /* Has password */
            {
              ++p; /* Skip the : */
              gurl->password = g_strndup(p, inner_temp - p);
              --p; /* Decrement below for user */
            }

          gurl->user = g_strndup(temp, p - temp);
          p = inner_temp;
          ++p; /* Skip the @ */
        }
      else
        p = temp;

      /* Hostname part */
      temp = p;
      while (*p != 0 && *p != ':' && *p != '/')
        ++p;

      gurl->hostname = g_strndup(temp, p - temp);
          
      if (*p == ':') /* Has port */
        {
          ++p;
          gurl->port = atoi(p);

          while (*p != 0 && *p != '/')
            ++p;
        }
    }
  
  /* Resource part */

  temp = p;

  while (*p != 0 && *p != '?' && *p != '#')
    ++p;

  if (p - temp > 0)
    {
      gurl->resource = g_strndup(temp, p - temp);
    }
  
  /* Query part */
  
  if (*p == '?')
    {
      temp = p;

      while (*p != 0 && *p != '#')
        ++p;

      ++temp; /* Skip the ? */
      gurl->query = g_strndup(temp, p - temp);
    }

  /* Fragment part */
  
  if (*p == '#')
    {
      ++p;
      gurl->fragment = g_strdup(p);
    }

  g_free (local_url);
  return gurl;
}


GURL*     
gnet_url_new_fields (const gchar* protocol, const gchar* hostname,
		const gint port, const gchar* resource, const gchar* user,
		const gchar* password, const gchar* query, const gchar* fragment)
{
  GURL* url = NULL;

  url = g_new0 (GURL, 1);
  if (protocol)		url->protocol = g_strdup (protocol);
  if (hostname)		url->hostname = g_strdup (hostname);
  url->port = port;
  if (resource)		url->resource = g_strdup (resource);
  if (user)		url->user = g_strdup (user);
  if (password)		url->password = g_strdup (password);
  if (query)		url->query = g_strdup (query);
  if (fragment)		url->fragment = g_strdup (fragment);
  
  return url;
}


GURL*     
gnet_url_clone (const GURL* url)
{
  GURL* url2;

  g_return_val_if_fail (url, NULL);

  url2 = g_new0 (GURL, 1);
  url2->protocol = g_strdup (url->protocol);
  url2->hostname = g_strdup (url->hostname);
  url2->port = url->port;
  url2->resource = g_strdup (url->resource);
  url2->user = g_strdup (url->user);
  url2->password = g_strdup (url->password);
  url2->query = g_strdup (url->query);
  url2->fragment = g_strdup (url->fragment);
  
  return url2;
}


void
gnet_url_delete (GURL* url)
{
  if (url)
    {
      g_free (url->protocol);
      g_free (url->hostname);
      g_free (url->resource);
      g_free (url->user);
      g_free (url->password);
      g_free (url->query);
      g_free (url->fragment);
      g_free (url);
    }
}


guint
gnet_url_hash (gconstpointer p)
{
  const GURL* url = (const GURL*) p;
  guint h = 0;

  g_return_val_if_fail (url, 0);

  if (url->protocol)	h =  g_str_hash (url->protocol);
  if (url->hostname)	h ^= g_str_hash (url->hostname);
  h ^= url->port;
  if (url->resource)	h ^= g_str_hash (url->resource);
  if (url->user)	h ^= g_str_hash (url->user);
  if (url->password)	h ^= g_str_hash (url->password);
  if (url->query)	h ^= g_str_hash (url->query);
  if (url->fragment)	h ^= g_str_hash (url->fragment);
  return h;
}


#define SAFESTRCMP(A,B) ((!(A) && !(B)) || ((A) && (B) && !strcmp((A),(B))))

gint
gnet_url_equal (gconstpointer p1, gconstpointer p2)
{
  const GURL* url1 = (const GURL*) p1;
  const GURL* url2 = (const GURL*) p2;

  g_return_val_if_fail (url1, 0);
  g_return_val_if_fail (url2, 0);

  if (url1->port == url2->port &&
      SAFESTRCMP(url1->protocol, url2->protocol) &&
      SAFESTRCMP(url1->hostname, url2->hostname) &&
      SAFESTRCMP(url1->resource, url2->resource) &&
      SAFESTRCMP(url1->user, url2->user) &&
      SAFESTRCMP(url1->password, url2->password) &&
      SAFESTRCMP(url1->query, url2->query) &&
      SAFESTRCMP(url1->fragment, url2->fragment))
    return 1;

  return 0;
}


void
gnet_url_set_protocol (GURL* url, const gchar* protocol)
{
  g_return_if_fail (url);

  if (url->protocol)
    {
      g_free (url->protocol);
      url->protocol = NULL;
    }

  if (protocol)
    url->protocol = g_strdup (protocol);
}


void
gnet_url_set_hostname (GURL* url, const gchar* hostname)
{
  g_return_if_fail (url);

  if (url->hostname)
    {
      g_free (url->hostname);
      url->hostname = NULL;
    }

  if (hostname)
    url->hostname = g_strdup (hostname);
}


void	
gnet_url_set_port (GURL* url, gint port)
{
  url->port = port;
}


void
gnet_url_set_resource (GURL* url, const gchar* resource)
{
  g_return_if_fail (url);

  if (url->resource)
    {
      g_free (url->resource);
      url->resource = NULL;
    }

  if (resource)
    {
      if (resource[0] == '/')
	url->resource = g_strdup (resource);
      else
	url->resource = g_strconcat ("/", resource, NULL);
    }
}

void
gnet_url_set_user (GURL* url, const gchar* user)
{
  g_return_if_fail (url);

  if (url->user)
    {
      g_free (url->user);
      url->user = NULL;
    }

  if (user)
    url->user = g_strdup (user);
}

void
gnet_url_set_password (GURL* url, const gchar* password)
{
  g_return_if_fail (url);

  if (url->password)
    {
      g_free (url->password);
      url->password = NULL;
    }

  if (password)
    url->password = g_strdup (password);
}

void
gnet_url_set_query (GURL* url, const gchar* query)
{
  g_return_if_fail (url);

  if (url->query)
    {
      g_free (url->query);
      url->query = NULL;
    }

  if (query)
    url->query = g_strdup (query);
}

void
gnet_url_set_fragment (GURL* url, const gchar* fragment)
{
  g_return_if_fail (url);

  if (url->fragment)
    {
      g_free (url->fragment);
      url->fragment = NULL;
    }

  if (fragment)
    url->fragment = g_strdup (fragment);
}

gchar*
gnet_url_get_nice_string (const GURL* url)
{
  gchar* rv = NULL;
  gchar* resource = NULL;
  GString* buffer = NULL;
  
  g_return_val_if_fail (url, NULL);
  g_return_val_if_fail (url->hostname, NULL);

  if (url->resource)
    {
      if (*url->resource == '/')
	resource = g_strdup (url->resource);
      else
	resource = g_strconcat ("/", url->resource, NULL);
    }
  else
    resource = g_strdup ("");

  buffer = g_string_sized_new(16);

  if (url->protocol != NULL)
    {
      g_string_sprintfa (buffer, "%s://", url->protocol);
    }

  if (url->user)
    {
      buffer = g_string_append (buffer, url->user);

      if (url->password)
        {
          g_string_sprintfa (buffer, ":%s", url->password);
        }

      buffer = g_string_append_c (buffer, '@');
    }

  buffer = g_string_append (buffer, url->hostname); 

  if (url->port)
    {
      g_string_sprintfa (buffer, ":%d", url->port);
    }

  buffer = g_string_append (buffer, resource);

  if (url->query)
    {
      g_string_sprintfa (buffer, "?%s", url->query);
    }

  if (url->fragment)
    {
      g_string_sprintfa (buffer, "#%s", url->fragment);
    }
  
  g_free (resource);

  /* Free only GString not data contained, return the data instead */
  rv = buffer->str;
  g_string_free(buffer, FALSE); 
  return rv;
}
